    #!/bin/python

import numpy as np
from scipy.optimize import linear_sum_assignment
import io
from collections import namedtuple
import sys
import os
import os.path


mr = None
mh = None

def make_spkmap(aspk):
    spkmap = {}
    lspkmaps = []
    lspkids = []
    lspkunids = []
    spkcount = 0
    spkunmap = []
    for spk in aspk:
        lspkmap = {}
        lspkid = []
        lspkunid = {}
        for s in spk.speaker:
            if not s in lspkmap:
                if not s in spkmap:
                    spkunmap.append(s)
                    spkmap[s] = spkcount
                    lspkmap[s] = len(lspkid)
                    lspkunid[spkcount] = len(lspkid)
                    lspkid.append(spkcount)
                    spkcount += 1
                else:
                    sid = spkmap[s]
                    lspkmap[s] = len(lspkid)
                    lspkunid[sid] = len(lspkid)
                    lspkid.append(sid)
        lspkmaps.append(lspkmap)
        lspkids.append(lspkid)
        lspkunids.append(lspkunid)
    return spkmap, spkunmap, spkcount, lspkmaps, lspkids, lspkunids


def range_to_frontiers(rng):
    rng.sort()
    pos = 0
    while pos < len(rng) - 1:
        if rng[pos][1] >= rng[pos + 1][0]:
            rng[pos] = (rng[pos][0], max(rng[pos][1], rng[pos + 1][1]))
            rng.pop(pos + 1)
        else:
            pos = pos + 1
    front = []
    for r in rng:
        front.append(("n", r[0]))
        front.append(("p", r[1]))
    return front

def filter_frontier_on_uem(front, uem):
    uemi = 0
    fri = 0
    fo = []
    while uemi != len(uem.start_time) and fri != len(front):
        if uem.start_time[uemi] < front[fri][1]:
            if uem.end_time[uemi] >= front[fri][1]:
                if front[fri][0] != 'n':
                    if len(fo) == 0 or fo[-1][1] < uem.start_time[uemi]:
                        fo.append(('n', uem.start_time[uemi]))
                    fo.append((front[fri][0], front[fri][1]))
                else:
                    fo.append((front[fri][0], front[fri][1]))
                fri += 1
            else:
                if front[fri][0] != 'n':
                    if len(fo) == 0 or fo[-1][1] < uem.start_time[uemi]:
                        fo.append(('n', uem.start_time[uemi]))
                    fo.append((front[fri][0], uem.end_time[uemi]))
                uemi += 1
        else:
            fri += 1
    return fo

def filter_frontiers_on_uem(afront, auem):
    afo = []
    for aid, front in enumerate(afront):
        uem = auem[aid]
        fo = []
        for fr in front:
            fo.append(filter_frontier_on_uem(fr, uem))
        afo.append(fo)
    return afo

def filter_frontier_on_uem_a(afront, auem):
    afo = []
    for aid, front in enumerate(afront):
        afo.append(filter_frontier_on_uem(front, auem[aid]))
    return afo

def merge_two_frontiers(front1, front2, end1, end2):
    frontr = []
    pos1 = 0
    pos2 = 0
    while pos1 < len(front1) or pos2 < len(front2):
        ctime = (
            front1[pos1][1]
            if pos2 == len(front2)
            else front2[pos2][1]
            if pos1 == len(front1)
            else min(front1[pos1][1], front2[pos2][1])
        )
        mode1 = end1 if pos1 == len(front1) else front1[pos1][0]
        mode2 = end2 if pos2 == len(front2) else front2[pos2][0]
        frontr.append((mode1 + mode2, ctime))
        if pos1 != len(front1) and front1[pos1][1] == ctime:
            pos1 += 1
        if pos2 != len(front2) and front2[pos2][1] == ctime:
            pos2 += 1
    return frontr


def make_merge_frontier(ahyp_union, aref_union, aref_frontiers_collar):
    afrontr = []
    for aid, hyp_union in enumerate(ahyp_union):
        ref_union = aref_union[aid]
        ref_frontiers_collar = aref_frontiers_collar[aid]
        hr = merge_two_frontiers(hyp_union, ref_union, "n", "n")
        frontr = []
        for f in ref_frontiers_collar:
            frontr.append(merge_two_frontiers(hr, f, "nn", "n"))
        afrontr.append(frontr)
    return afrontr


def make_frontiers(aspk, aspkmap, aspkid):
    afront = []
    for aid, spk in enumerate(aspk):
        spkmap = aspkmap[aid]
        spkcount = len(aspkid[aid])
        rngs = [[] for i in range(spkcount)]
        for i in range(0, len(spk.speaker)):
            spki = spkmap[spk.speaker[i]]
            rngs[spki].append((spk.start_time[i], spk.end_time[i]))
        front = []
        for r in rngs:
            front.append(range_to_frontiers(r))
        afront.append(front)
    return afront


def make_union_frontiers(aspk):
    afo = []
    for spk in aspk:
        rngs = []
        for i in range(0, len(spk.speaker)):
            rngs.append((spk.start_time[i], spk.end_time[i]))
        afo.append(range_to_frontiers(rngs))
    return afo


def frontiers_add_collar(front, collar):
    cfront = []
    for f in front:
        a = f[1] - collar
        b = f[1] + collar
        if a < 0:
            a = 0
        if len(cfront) == 0 or a > cfront[-1][1]:
            cfront.append((f[0], a))
            cfront.append(("t", b))
        else:
            cfront[-1] = ("t", b)
    return cfront


def make_times(afront, alspkids, spkcount):
    times = [0] * spkcount
    for aid, front in enumerate(afront):
        for sid, s in enumerate(front):
            time = 0
            ptime = 0
            for p in s:
                if p[0] == "n":
                    ptime = p[1]
                elif p[0] == "p":
                    time += p[1] - ptime
            times[alspkids[aid][sid]] += time
    return times

def add_time(thyp, thyn, mode, eh, er, tc, efa, emiss, econf):
    if mode == "ppp":
        return eh, er + thyn, tc + thyp, efa, emiss, econf + thyn
    if mode == "ppn":
        return eh + thyp, er, tc, efa, emiss, econf
    if mode == "ppt":
        return eh, er, tc + thyp, efa, emiss, econf
    if mode == "pnn":
        return eh + thyp, er, tc, efa + thyp, emiss, econf
    if mode == "pnt":
        return eh, er, tc + thyp, efa, emiss, econf
    if mode == "npp":
        return eh, er + thyn, tc, efa, emiss + thyn, econf
    # npn npt nnn nnt
    return eh, er, tc, efa, emiss, econf


def compute_times(afrontr, aspkunidr, rid, afronth, aspkunidh, hid):
    aeh = 0
    aer = 0
    atc = 0
    aefa = 0
    aemiss = 0
    aeconf = 0
    for aid in range(len(afrontr)):
        spkunidr = aspkunidr[aid]
        spkunidh = aspkunidh[aid]
        frontr = afrontr[aid][spkunidr[rid]] if rid in spkunidr else [('n', 0.0)]
        fronth = afronth[aid][spkunidh[hid]] if hid in spkunidh else [('n', 0.0)]
        eh = 0
        er = 0
        tc = 0
        efa = 0
        emiss = 0
        econf = 0
        hpos = 0
        tbeg = 0
        thyp = 0
        hypbef = 0
        for f in frontr:
            tend = f[1]
            while hpos < len(fronth):
                dinter = min(fronth[hpos][1], tend)
                if fronth[hpos][0] == "p":
                    thyp += dinter - hypbef
                if fronth[hpos][1] > tend:
                    break
                hypbef = dinter
                hpos += 1
            eh, er, tc, efa, emiss, econf = add_time(
                thyp, tend - tbeg - thyp, f[0], eh, er, tc, efa, emiss, econf
            )
    
            if hpos < len(fronth):
                hypbef = min(fronth[hpos][1], tend)
            tbeg = tend
            thyp = 0
        while hpos < len(fronth):
            if fronth[hpos][0] == "p":
                thyp += fronth[hpos][1] - tbeg
            tbeg = fronth[hpos][1]
            hpos += 1
        eh, er, tc, efa, emiss, econf = add_time(
            thyp, 0, "pnn", eh, er, tc, efa, emiss, econf
        )
        aeh += eh
        aer += er
        atc += tc
        aefa += efa
        aemiss += emiss
        aeconf += econf
        
    return (
        round(aeh, 3),
        round(aer, 3),
        round(atc, 3),
        round(aefa, 3),
        round(aemiss, 3),
        round(aeconf, 3),
    )


def compute_miss(afunion, afront, alspkid, spkcount):
    miss = [0] * spkcount
    for aid, front in enumerate(afront):
        funion = afunion[aid]
        lspkid = alspkid[aid]
        for spkid, f1 in enumerate(front):
            hpos = 0
            tbeg = 0
            thyp = 0
            hypbef = 0
            fa = 0
            for f in funion:
                tend = f[1]
                while hpos < len(f1):
                    dinter = min(f1[hpos][1], tend)
                    if f1[hpos][0] == "p":
                        thyp += dinter - hypbef
                    if f1[hpos][1] > tend:
                        break
                    hypbef = dinter
                    hpos += 1
                if f[0] == "n":
                    fa += thyp
                if hpos < len(f1):
                    hypbef = min(f1[hpos][1], tend)
                tbeg = tend
                thyp = 0
            while hpos < len(f1):
                if f1[hpos][0] == "p":
                    thyp += f1[hpos][1] - tbeg
                tbeg = f1[hpos][1]
                hpos += 1
            fa += thyp
            fa = round(fa, 3)
            miss[lspkid[spkid]] += fa
    return miss


def accumulate_confusion(afref, afrefc, afhyp, map_rh, map_hr, ref_spkcount, hyp_spkcount, ref_lspkunids, hyp_lspkunids):
    correct_ref = [0] * ref_spkcount
    correct_hyp = [0] * hyp_spkcount
    lost_ref = [0] * ref_spkcount
    lost_hyp = [0] * hyp_spkcount
    confusion_matrix = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")

    for aid, fhyp in enumerate(afhyp):
        ref_lspkunid = ref_lspkunids[aid]
        hyp_lspkunid = hyp_lspkunids[aid]
        fri = [0] * ref_spkcount
        fhi = [0] * hyp_spkcount
        fref1 = afref[aid]
        fref1c = afrefc[aid]
        frefx = [None] * ref_spkcount
        for i in range(ref_spkcount):
            if i in ref_lspkunid:
                sid = ref_lspkunid[i]
                frefx[i] = fref1c[sid] if map_rh[i] in hyp_lspkunid else fref1[sid]
        fhypx = [None] * hyp_spkcount
        for i in range(hyp_spkcount):
            if i in hyp_lspkunid:
                sid = hyp_lspkunid[i]
                fhypx[i] = fhyp[sid]
        
        cur_time = 0
        while True:
            ridx = []
            r_is_t = []
            hidx = []
            time = -1
    
            # Build the list of who is in the segment
            for i in range(ref_spkcount):
                if frefx[i] != None and fri[i] != len(frefx[i]):
                    cf = frefx[i][fri[i]]
                    if time == -1 or cf[1] < time:
                        time = cf[1]
                    if cf[0] != "n":
                        ridx.append(i)
                        r_is_t.append(cf[0] == "t")
    
            for i in range(hyp_spkcount):
                if fhypx[i] != None and fhi[i] != len(fhypx[i]):
                    cf = fhypx[i][fhi[i]]
                    if time == -1 or cf[1] < time:
                        time = cf[1]
                    if cf[0] != "n":
                        hidx.append(i)
    
            if time == -1:
                break
    
            # Only do the computations when there's something to do
            if len(ridx) > 0 or len(hidx) > 0:
                duration = time - cur_time
    
                # Hyp and ref mapped together end up in correct time and are removed from the lists
                i = 0
                while i != len(ridx):
                    r = ridx[i]
                    h = map_rh[r]
                    dropped = False
                    if h != -1:
                        slot = -1
                        for j in range(len(hidx)):
                            if hidx[j] == h:
                                slot = j
                                break
                        if slot != -1:
                            correct_ref[r] += duration
                            correct_hyp[h] += duration
                            ridx.pop(i)
                            r_is_t.pop(i)
                            hidx.pop(slot)
                            dropped = True
                    if not dropped:
                        i += 1
    
                # Ref in transition is removed from the list if mapped to some hyp
                i = 0
                while i != len(ridx):
                    r = ridx[i]
                    if r_is_t[i] and map_rh[r] != -1:
                        ridx.pop(i)
                        r_is_t.pop(i)
                    else:
                        i += 1
    
                if len(hidx) == 0:
                    # If there's no hyp, we're all in lost_ref
                    for r in ridx:
                        lost_ref[r] += duration
    
                elif len(ridx) == 0:
                    # If there's no ref, we're all in lost_hyp
                    for h in hidx:
                        lost_hyp[h] += duration
    
                else:
                    # Otherwise we're in confusion.  Amount of confusion time to give
                    # is equal to the max of the ref and hyp times
                    conf_time = max(len(ridx), len(hidx)) * duration
    
                    # Number of slots, otoh, is equal to the product of the number of
                    # refs and hyps
                    conf_slots = len(ridx) * len(hidx)
    
                    # Give the time equally in all slots
                    conf_one_time = conf_time / conf_slots
                    for r in ridx:
                        for h in hidx:
                            confusion_matrix[r, h] += conf_one_time
    
            # Step all the done segments
            for r in range(ref_spkcount):
                if frefx[r] != None and fri[r] != len(frefx[r]) and frefx[r][fri[r]][1] == time:
                    fri[r] += 1
            for h in range(hyp_spkcount):
                if fhypx[h] != None and fhi[h] != len(fhypx[h]) and fhypx[h][fhi[h]][1] == time:
                    fhi[h] += 1
            cur_time = time

    return correct_ref, correct_hyp, lost_ref, lost_hyp, confusion_matrix


def compute_der(aref, ahyp, auem, map, collar):
    ref_spkmap, ref_spkunmap, ref_spkcount, ref_lspkmaps, ref_lspkids, ref_lspkunids = make_spkmap(aref)
    hyp_spkmap, hyp_spkunmap, hyp_spkcount, hyp_lspkmaps, hyp_lspkids, hyp_lspkunids = make_spkmap(ahyp)

    global mr, mh
    mr = ref_spkunmap
    mh = hyp_spkunmap

    aref_frontiers = filter_frontiers_on_uem(make_frontiers(aref, ref_lspkmaps, ref_lspkids), auem)
    ahyp_frontiers = filter_frontiers_on_uem(make_frontiers(ahyp, hyp_lspkmaps, hyp_lspkids), auem)
    aref_frontiers_collar = []
    for aid, ref_frontiers in enumerate(aref_frontiers):
        ref_frontiers_collar = []
        for front in ref_frontiers:
            ref_frontiers_collar.append(filter_frontier_on_uem(frontiers_add_collar(front, collar), auem[aid]))
        aref_frontiers_collar.append(ref_frontiers_collar)

    aref_union = filter_frontier_on_uem_a(make_union_frontiers(aref), auem)
    ahyp_union = filter_frontier_on_uem_a(make_union_frontiers(ahyp), auem)

    amerge_frontiers = make_merge_frontier(ahyp_union, aref_union, aref_frontiers_collar)

    ref_times = make_times(aref_frontiers, ref_lspkids, ref_spkcount)
    hyp_times = make_times(ahyp_frontiers, hyp_lspkids, hyp_spkcount)

    eh = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    er = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    tc = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    efa = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    emiss = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    econf = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    de = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")

    miss_hyp = compute_miss(aref_union, ahyp_frontiers, hyp_lspkids, hyp_spkcount)
    miss_ref = compute_miss(ahyp_union, aref_frontiers, ref_lspkids, ref_spkcount)
    for r in range(ref_spkcount):
        for h in range(hyp_spkcount):
            (
                eh[r, h],
                er[r, h],
                tc[r, h],
                efa[r, h],
                emiss[r, h],
                econf[r, h],
            ) = compute_times(amerge_frontiers, ref_lspkunids, r, ahyp_frontiers, hyp_lspkunids, h)
            de[r, h] = ref_times[r] + miss_hyp[h] - efa[r, h] - emiss[r, h] - econf[r, h]
    newspkmap = {}
    map_rh = [-1] * ref_spkcount
    map_hr = [-1] * hyp_spkcount
    solved_ref = [0] * ref_spkcount
    solved_hyp = [0] * hyp_spkcount
    for rs in map:
        hs = map[rs]
        newspkmap[rs] = hs
        rid = ref_spkmap[rs] if rs in ref_spkmap else -1
        hid = hyp_spkmap[hs] if hs in hyp_spkmap else -1
        if rid != -1:
            solved_ref[rid] = -1
        if hid != -1:
            solved_hyp[hid] = -1
        if rid != -1 and hid != -1:
            map_rh[rid] = hid
            map_hr[hid] = rid

    solved_ref_back = []
    solved_hyp_back = []
    for i in range(0, ref_spkcount):
        if solved_ref[i] == 0:
            solved_ref[i] = len(solved_ref_back)
            solved_ref_back.append(i)
    for i in range(0, hyp_spkcount):
        if solved_hyp[i] == 0:
            solved_hyp[i] = len(solved_hyp_back)
            solved_hyp_back.append(i)

    solved_ref_count = len(solved_ref_back)
    solved_hyp_count = len(solved_hyp_back)
    opt_size = max(solved_ref_count, solved_hyp_count)
    costs = np.zeros((opt_size, opt_size), dtype="float64")
    for r1 in range(solved_ref_count):
        r = solved_ref_back[r1]
        for h1 in range(solved_hyp_count):
            h = solved_hyp_back[h1]
            costs[r1, h1] = -round(de[r, h] * 1000)            

    (map1, map2) = linear_sum_assignment(costs)
    for i1 in range(0, opt_size):
        i = map1[i1]
        j = map2[i1]
        if (
            i < solved_ref_count
            and j < solved_hyp_count
            and de[i, j] > 0
            and tc[i, j] > 0
        ):
            r = solved_ref_back[i]
            h = solved_hyp_back[j]
            map_rh[r] = h
            map_hr[h] = r

    for i, j in enumerate(map_rh):
        if j != -1:
            newspkmap[ref_spkunmap[i]] = hyp_spkunmap[j]

    (
        correct_ref,
        correct_hyp,
        lost_ref,
        lost_hyp,
        confusion_matrix,
    ) = accumulate_confusion(aref_frontiers, aref_frontiers_collar, ahyp_frontiers, map_rh, map_hr, ref_spkcount, hyp_spkcount, ref_lspkunids, hyp_lspkunids)

    conf = 0
    for r in range(ref_spkcount):
        for h in range(hyp_spkcount):
            conf += confusion_matrix[r, h]
    totaltime = 0
    miss = 0
    for r in range(ref_spkcount):
        totaltime += ref_times[r]
        miss += lost_ref[r]
    fa = 0
    for h in range(hyp_spkcount):
        fa += lost_hyp[h]

    return fa + miss + conf, totaltime, newspkmap

Spki = namedtuple('Spki', ['speaker', 'start_time', 'end_time'])
Uemi = namedtuple('Uemi', ['start_time', 'end_time'])

def load_mdtm(file):
    global Spki
    res = Spki(speaker=[], start_time=[], end_time=[])
    for l in io.open(file):
        e = l.split()
        res.start_time.append(round(float(e[2]), 3))
        res.end_time.append(round(float(e[2]) + float(e[3]), 3))
        res.speaker.append(e[7])
    return res

def load_uem(file):
    global Uemi
    res = Uemi(start_time=[], end_time=[])
    for l in io.open(file):
        e = l.split()
        res.start_time.append(round(float(e[2]), 3))
        res.end_time.append(round(float(e[3]), 3))
    return res

    
def load_map(file):
    res = {}
    for l in io.open(file):
        e = l.split()
        res[e[0]] = e[1]
    return res

def find_file(pdir, names, ext):
    for i in range(0, 2):
        p = pdir + '/' + names[i] + ext
        if os.path.isfile(p):
            return p
    err = ""
    for i in range(0, 2):
        if i == 0:
            err += "Neither "
        else:
            err += " nor "
        err += pdir + '/' + names[i] + ext
    err += ' war found.'
    print(err, file = sys.stderr)
    sys.exit(1)


def der_cross(lifelong_list, ref_mdtm_dir, ref_uem_dir, hyp_mdtm_dir):

    with open(lifelong_list, "r") as fh:
       llist = [l.rstrip() for l in fh.readlines()]

    map = {}

    total_error = 0
    total_time = 0

    ref = []
    uem = []
    hyp = []

    for l in open(lifelong_list):
        e = l.split()
        ref.append(load_mdtm(find_file(ref_mdtm_dir, e, ".mdtm")))
        uem.append(load_uem (find_file(ref_uem_dir, e, ".uem")))
        hyp.append(load_mdtm(find_file(hyp_mdtm_dir, e, ".mdtm")))


        #print("%s: DER: %6.2f%%  Time: %.2fs" % (e[0], der, time))

    (error, time, map) = compute_der(ref, hyp, uem, map, 0.250)

    print("Global DER: %6.2f%%  Total time: %.2fs" % (error * 100.0 / time, time))
    
    #mapf = open(sys.argv[5], 'w')
    #for s in map:
    #    print("%s %s" % (s, map[s]), file=mapf)



#if __name == "__main__":
#
#    if len(sys.argv) != 6:
#        print("Computes the DER a series of reference/hypothesis and output the final mapping file.\nUsage:\n  %s dev.lst ref-mdtm-dir ref-uem-dir hyp-mdtm-dir output-map.txt\n" % sys.argv[0])
#        sys.exit(0)
#    der_cross(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])




#!/usr/bin/python




def apply_correction(hyp, scores, c1, c2, method="min"):
    """

    """
    if method == "min":
        hyp = apply_correction_min(hyp, scores, c1, c2)
    elif method == "max":
        hyp = apply_correction_max(hyp, scores, c1, c2)
    elif method == "average":
        hyp = apply_correction_avr(hyp, scores, c1, c2)

    return hyp


def apply_correction_min(hyp,scores,c1,c2):
    c1list=[]
    c2list=[]

    for i in range(len(hyp)):
        if i!=c1 and hyp[i]['cluster'] == hyp[c1]['cluster']:
            c1list.append(i)
        if i!=c2 and hyp[i]['cluster'] == hyp[c2]['cluster']:
            c2list.append(i)

    if len(c1list)==0 and len(c2list)!=0:
        c2_to_c1 = True

    elif len(c2list)==0 and len(c1list)!=0:
        c2_to_c1 = False

    elif len(c1list)==0 and len(c2list)==0:
        c2_to_c1 = True

    else:
        c1dist = []
        c2dist = []
        for i in range(len(c1list)):
            c1dist.append(scores.scoremat[c1][c1list[i]])
        for i in range(len(c2list)):
            c2dist.append(scores.scoremat[c2][c2list[i]])
        if min(c2dist) <  min(c1dist):
            c2_to_c1 = True
        else:
            c2_to_c1 = False

    if c2_to_c1:
        hyp[c1]['cluster'] = hyp[c2]['cluster']
    else:
        hyp[c2]['cluster'] = hyp[c1]['cluster']
    return hyp




def apply_correction_max(hyp,scores,c1,c2):
    c1list=[]
    c2list=[]

    for i in range(len(hyp)):
        if i!=c1 and hyp[i]['cluster'] == hyp[c1]['cluster']:
            c1list.append(i)
        if i!=c2 and hyp[i]['cluster'] == hyp[c2]['cluster']:
            c2list.append(i)

    if len(c1list)==0 and len(c2list)!=0:
        c2_to_c1 = True

    elif len(c2list)==0 and len(c1list)!=0:
        c2_to_c1 = False

    elif len(c1list)==0 and len(c2list)==0:
        c2_to_c1 = True

    else:
        c1dist = []
        c2dist = []
        for i in range(len(c1list)):
            c1dist.append(scores.scoremat[c1][c1list[i]])
        for i in range(len(c2list)):
            c2dist.append(scores.scoremat[c2][c2list[i]])
        if max(c2dist) <  max(c1dist):
            c2_to_c1 = True
        else:
            c2_to_c1 = False

    if c2_to_c1:
        hyp[c1]['cluster'] = hyp[c2]['cluster']
    else:
        hyp[c2]['cluster'] = hyp[c1]['cluster']
    return hyp

def apply_correction_avr(hyp,scores,c1,c2):
    c1list=[]
    c2list=[]

    for i in range(len(hyp)):
        if i!=c1 and hyp[i]['cluster'] == hyp[c1]['cluster']:
            c1list.append(i)
        if i!=c2 and hyp[i]['cluster'] == hyp[c2]['cluster']:
            c2list.append(i)

    if len(c1list)==0 and len(c2list)!=0:
        c2_to_c1 = True

    elif len(c2list)==0 and len(c1list)!=0:
        c2_to_c1 = False

    elif len(c1list)==0 and len(c2list)==0:
        c2_to_c1 = True

    else:
        c1dist = []
        c2dist = []
        for i in range(len(c1list)):
            c1dist.append(scores.scoremat[c1][c1list[i]])
        for i in range(len(c2list)):
            c2dist.append(scores.scoremat[c2][c2list[i]])
        if sum(c2dist)/float(len(c2dist)) <  sum(c1dist)/float(len(c1dist)):
            c2_to_c1 = True
        else:
            c2_to_c1 = False

    if c2_to_c1:
        hyp[c1]['cluster'] = hyp[c2]['cluster']
    else:
        hyp[c2]['cluster'] = hyp[c1]['cluster']
    return hyp





from .database import Database

from .user_simulation import UserSimulation
from .user_simulation import MessageToUser 
from .user_simulation import Request

from .der_single import der_single_averaged
from .der_cross import der_cross

from .lium_baseline.system import allies_load_seg
from .lium_baseline.system import allies_initial_training
from .lium_baseline.system import lium_iv_initial_training
from .lium_baseline.system import lium_xv_initial_training
from .lium_baseline.system import extract_vectors
from .lium_baseline.system import allies_init_seg
from .lium_baseline.system import allies_within_show_HAL
from .lium_baseline.system import allies_cross_show_clustering
from .lium_baseline.system import allies_write_diar

from .lium_baseline.interactive import apply_correction
from .lium_baseline.interactive import apply_correction_min
from .lium_baseline.interactive import apply_correction_max
from .lium_baseline.interactive import apply_correction_avr

__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2020-2021 Anthony Larcher"
__license__ = "LGPL"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'
__version__="0.0.1"



import copy
import decimal
import fastcluster
import h5py
import logging
import numpy
import os
import pandas
import pickle
import sidekit
import soundfile
import s4d
import sys
import torch
import yaml

from scipy.cluster import hierarchy as hac

from scipy.spatial.distance import squareform

from .utils import add_show_in_id
from .utils import remove_show_from_id
from .utils import s4d_to_allies

from .utils import keep_recurring_speakers
#from .incremental import perform_iv_seg
#from .incremental import cross_show

from ..user_simulation import MessageToUser
from ..user_simulation import Request

from .interactive import apply_correction

#from s4d import viterbi, segmentation
#from s4d.clustering import hac_bic
#from s4d.model_iv import ModelIV



def concat_statservers(ss_1, ss_2):
    out_sessions = numpy.unique(ss_1.modelset).shape[0]

    in_sessions = numpy.unique(ss_2.modelset).shape[0]

    tot_sessions = ss_1.modelset.shape[0] + ss_2.modelset.shape[0]

    flt_iv = sidekit.StatServer()
    flt_iv.stat0 = numpy.ones((tot_sessions, 1))
    flt_iv.stat1 = numpy.ones((tot_sessions, ss_1.stat1.shape[1]))
    flt_iv.modelset = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.segset = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.start = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.stop = numpy.empty(tot_sessions, dtype="|O")

    for idx in range(ss_1.modelset.shape[0]):
        flt_iv.stat1[idx, :] = ss_1.stat1[idx, :]
        flt_iv.modelset[idx] = str(ss_1.modelset[idx])
        flt_iv.segset[idx] = ss_1.segset[idx]
        flt_iv.start[idx] = ss_1.start[idx]
        flt_iv.stop[idx] = ss_1.stop[idx]

    for idx2 in range(ss_2.modelset.shape[0]):
        idx = idx2 + ss_1.modelset.shape[0]
        flt_iv.stat1[idx, :] = ss_2.stat1[idx2, :]
        flt_iv.modelset[idx] = str(ss_2.modelset[idx2])
        flt_iv.segset[idx] = ss_2.segset[idx2]
        flt_iv.start[idx] = ss_2.start[idx2]
        flt_iv.stop[idx] = ss_2.stop[idx2]

    return flt_iv


def rename_models(within_iv, within_diar, existing_models):                                                                                                                                                                               
    """
    """
    within_models = dict()
    idx = 0

    for mod in set(within_iv.modelset):
        while f"Spk{idx}" in existing_models:
            idx += 1
        within_models[mod] = f"Spk{idx}"
        idx += 1

    within_iv.modelset = within_iv.modelset.astype('U50')
    for idx, mod in enumerate(within_iv.modelset):
        within_iv.modelset[idx] = within_models[mod]

    for idx, seg in enumerate(within_diar.segments):
        within_diar.segments[idx]['cluster'] = within_models[seg['cluster']]

    return within_iv, within_diar


def perform_iv_seg(sn_mean,
                   sn_cov,
                   plda_mean,
                   plda_f,
                   plda_sigma,
                   th_W,
                   initial_diar,
                   within_iv,
                   hac_method,
                   precomputed_scores=None):
    """

    """
    within_iv_mean = within_iv.mean_stat_per_model()

    # Normalize i-vectors
    within_iv_mean.spectral_norm_stat1(sn_mean[:1], sn_cov[:1])
    if within_iv_mean.modelset.shape[0] > 1:
        if precomputed_scores is None:
            # Compute scores
            ndx = sidekit.Ndx(models=within_iv_mean.modelset, testsegs=within_iv_mean.modelset)
            scores = sidekit.iv_scoring.fast_PLDA_scoring(within_iv_mean, within_iv_mean, ndx,
                                                          plda_mean,
                                                          plda_f,
                                                          plda_sigma,
                                                          p_known=0.0,
                                                          scaling_factor=1.0,
                                                          check_missing=False)
            scores.scoremat = 0.5 * (scores.scoremat + scores.scoremat.transpose())
        else:
            scores = precomputed_scores

        # Run HAC clustering
        new_diar, cluster_dict, merge = s4d.clustering.hac_iv.hac_iv(initial_diar, scores, threshold=th_W, method=hac_method)

        # Update the model names of i-vector
        # reverse the cluster dict:
        update_dict = dict()
        for k, v in cluster_dict.items():
            for _v in v:
                update_dict[_v] = k
        
        for idx, mod in enumerate(within_iv.modelset):
            if mod in update_dict:
                within_iv.modelset[idx] = update_dict[mod]

    else:
        new_diar = initial_diar

    return new_diar, within_iv, scores


def cross_show(model, model_cfg, previous_iv, previous_diar, within_iv, within_diar, th_x, incremental=True):
    """
    """
    # Make sure model names are not overlapping (change within show models if needed)
    within_iv, within_diar = rename_models(within_iv, within_diar, existing_models=previous_iv.modelset)
    # backup previous_iv and within_iv
    within_iv_backup = copy.deepcopy(within_iv)
    previous_iv_backup = copy.deepcopy(previous_iv)

    # get the mean_per_model for previous and within
    within_iv_mean = within_iv.mean_stat_per_model()
    previous_iv_mean = previous_iv.mean_stat_per_model()

    #print(f"nombre de iv {within_iv.stat1.shape}, et de moyennes: {within_iv_mean.stat1.shape}")

    # merge the mean_per_model for previous and within
    ll_iv_mean = concat_statservers(previous_iv_mean, within_iv_mean)

    # compute scores, constrain and run HAC clustering
    # Normalize all
    ll_iv_mean.spectral_norm_stat1(model["model_iv"].sn_mean[:1], model["model_iv"].sn_cov[:1])

    # Compute the score matrix
    ndx = sidekit.Ndx(models=ll_iv_mean.modelset, testsegs=ll_iv_mean.modelset)
    scores = sidekit.iv_scoring.fast_PLDA_scoring(ll_iv_mean, ll_iv_mean, ndx,
                                                  model["model_iv"].plda_mean,
                                                  model["model_iv"].plda_f,
                                                  model["model_iv"].plda_sigma,
                                                  p_known=0.0,
                                                  scaling_factor=1.0,
                                                  check_missing=False)
    scores.scoremat = -0.5 * (scores.scoremat + scores.scoremat.transpose())

    # Constrain the scores to forbid any new clustering between previous shows
    lowest_distance = numpy.min(scores.scoremat)
    for iv_idx in range(previous_iv_mean.modelset.shape[0]):
        for iv_jdx in range(previous_iv_mean.modelset.shape[0]):
            if previous_iv_mean.modelset[iv_idx] == previous_iv_mean.modelset[iv_jdx]:
                scores.scoremat[iv_idx, iv_jdx] = lowest_distance - 1
            else:
                scores.scoremat[iv_idx, iv_jdx] = numpy.inf


    # Add to keep the within show clustering
    for iv_idx in range(previous_iv_mean.modelset.shape[0], ll_iv_mean.modelset.shape[0]):
        for iv_jdx in range(previous_iv_mean.modelset.shape[0], ll_iv_mean.modelset.shape[0]):
            if ll_iv_mean.modelset[iv_idx] == ll_iv_mean.modelset[iv_jdx]:
                scores.scoremat[iv_idx, iv_jdx] = lowest_distance - 1
            else:
                scores.scoremat[iv_idx, iv_jdx] = numpy.inf
    # fin modif

    # Modify the scores between old shows and the current one accordingly
    numpy.fill_diagonal(scores.scoremat, numpy.inf)
    for idx in range(previous_iv_mean.modelset.shape[0], ll_iv_mean.modelset.shape[0]):
        for jdx in range(previous_iv_mean.modelset.shape[0]):
            scores.scoremat[idx, jdx] = numpy.max(scores.scoremat[idx, numpy.where(previous_iv_mean.modelset == previous_iv_mean.modelset[jdx])])
            scores.scoremat[jdx, idx] = scores.scoremat[idx, jdx]
    numpy.fill_diagonal(scores.scoremat, 0)

    numpy.fill_diagonal(scores.scoremat, numpy.inf)

    #with open("matrix_ cross_show.txt", 'w') as fh:
    #    for ii in range(scores.scoremat.shape[0]):
    #        for jj in range(scores.scoremat.shape[1]):
    #            fh.write(f"{scores.scoremat[ii, jj]:3.1f},\t")
    #        fh.write("\n")

    mini0 = numpy.min(scores.scoremat) - 1
    scores.scoremat -= mini0

    modelset_seg_idx = dict()
    for seg in previous_diar.segments:
        modelset_seg_idx[seg['cluster']] = numpy.where(ll_iv_mean.modelset == seg['cluster'])[0]

    numpy.fill_diagonal(scores.scoremat, 0.0)
    squareform_plda = squareform(scores.scoremat)
    #Z = fastcluster.linkage(squareform_plda, method='complete', metric='euclidean', preserve_input=True)
    Z = fastcluster.linkage(squareform_plda, method='complete', preserve_input=True)

    """
    if we want to set the final number of cluster (approach DiHard)
    the next line should be replaced with

    final_cluster_nb = 12
    T = hac.fcluster(Z, t=final_cluster_nb, criterion="maxclust")

    Beware, final_cluster_nb should be higher than than the number of cluster existing in previous_iv_mean
    """
    T = hac.fcluster(Z, th_x - mini0, 'distance')

    # DSon't allow to modify the names of previously existing clusters
    # Create a dictionnary with old_model_name as key and new_luster as value
    cluster_dict = dict()
    clusters_by_index = dict()
    for ii in range(T.shape[0]):
        if T[ii] not in clusters_by_index:
            clusters_by_index[T[ii]] = ll_iv_mean.modelset[ii]
        cluster_dict[ll_iv_mean.modelset[ii]] = clusters_by_index[T[ii]]

    #print("merge dictionnary")
    #for k, v in cluster_dict.items():
    #    if not v == k:
    #        print(f"{k}\t{v}")

    # concatenere previous_iv et within_iv
    new_previous_iv = concat_statservers(previous_iv_backup, within_iv_backup)
    new_previous_diar = copy.deepcopy(previous_diar)
    new_previous_diar.segments += within_diar.segments

    # Modify the model names for i-vectors
    for ii, mod in enumerate(new_previous_iv.modelset):
        new_previous_iv.modelset[ii] = cluster_dict[mod]

    # modify the model names for diar
    for ii, seg in enumerate(new_previous_diar.segments):
        new_previous_diar.segments[ii]['cluster'] = cluster_dict[seg['cluster']]

    for ii, seg in enumerate(within_diar.segments):
        if within_diar.segments[ii]['cluster'] != cluster_dict[seg['cluster']]:
            print(f"change { within_diar.segments[ii]['cluster']} to {cluster_dict[seg['cluster']]}")
        within_diar.segments[ii]['cluster'] = cluster_dict[seg['cluster']]

    return new_previous_iv, new_previous_diar, within_diar

def allies_load_seg(model_cfg, th, show):

    current_diar = s4d.Diar.read_seg(f"{model_cfg['tmp_dir']}/seg/within_{th}/{show}.seg")
    current_vec = sidekit.StatServer(f"{model_cfg['tmp_dir']}/seg/within_{th}/{show}_vec.h5")

    return current_diar, current_vec


def allies_initial_training(system_config, train_data):
    """
    Train the initial Diarization system
    """
    # Load the config file
    with open(system_config, 'r') as fh:
        model_cfg = yaml.load(fh, Loader=yaml.FullLoader)

    if model_cfg["model"]["type"] == "lium_iv":
        model, model_cfg = lium_iv_initial_training(model_cfg, train_data)

    elif model_cfg["model"]["type"] == "lium_xv":
       model, model_cfg = lium_xv_initial_training(model_cfg, train_data)

    elif model_cfg["model"]["type"] == "lium_sliding_iv":
       model = None

    elif model_cfg["model"]["type"] == "lium_sliding_xv":
        model = None

    return model, model_cfg


def lium_iv_initial_training(model_cfg, train_data):
    """
    Extract features and store them in tmp folder
    Train UBM
    Train TV
    Extract i-vectors
    Train PLDA

    :param model_cfg: a dictionnary describing the system parameters 
    :param train_data: a Data generator used to access the data
    """
    fe = sidekit.FeaturesExtractor(audio_filename_structure="",
                                   feature_filename_structure="feat/{}.h5",
                                   sampling_frequency=16000,
                                   lower_frequency=133.3333,
                                   higher_frequency=6855.4976,
                                   filter_bank="log",
                                   filter_bank_size=40,
                                   window_size=0.025,
                                   shift=0.01,
                                   ceps_number=13,
                                   pre_emphasis=0.97,
                                   keep_all_features=True,
                                   vad='percentil',
                                   save_param=["energy", "cep", "vad"],
                                   compressed='none')

    fs_seg = sidekit.FeaturesServer(feature_filename_structure=model_cfg['tmp_dir']+ "/feat/{}.h5",
                                    dataset_list = ['energy', 'cep'],
                                    keep_all_features=True,
                                    mask='[1-13]',
                                    delta=False,
                                    double_delta=False)


    fs_acoustic = sidekit.FeaturesServer(feature_filename_structure=model_cfg['tmp_dir'] + "/feat/{}.h5",
                                         dataset_list = ['energy', 'cep'],
                                         keep_all_features=False,
                                         delta=True,
                                         double_delta=True,
                                         feat_norm='cmvn_sliding')

    train_diar = s4d.diar.Diar()
    train_diar.add_attribut(new_attribut='gender', default='U')
    name_dict = {}
    # iterate on train_data
    if not os.path.isdir(f"{model_cfg['tmp_dir']}/feat"):
        os.makedirs(f"{model_cfg['tmp_dir']}/feat")

    for idx, (show, file_info, uem, ref, filename) in enumerate(train_data):

        if not os.path.isfile(f"{model_cfg['tmp_dir']}/feat/{show}.h5"):
            print(f"\t Extract features: {show}")
            fe.save(show,
                    channel=0,
                    input_audio_filename=filename,
                    output_feature_filename=f"{model_cfg['tmp_dir']}/feat/{show}.h5")

        # Use the ref info to fill the Diar object
        for spk, start, end in zip(ref.speaker, ref.start_time, ref.end_time):
            train_diar.append(show=show,
                              cluster= spk,
                              start=int(round(float(start) * 100.)),
                              stop=int(round(float(end) * 100.))
                              #stop=int(round(float(end) * 100.)) + int(round(float(start) * 100.))
                              )

    print(f'Training data contains:')
    print("\t" + f'{len(train_diar.unique("show"))} files')
    print("\t" + f'{len(train_diar.unique("cluster"))} speakers')
    print("\t" + f'{len(train_diar)} segments')
    name_dict[show] = int(idx)


    ######################################
    # Train UBM
    ######################################
    ubm_diar = copy.deepcopy(train_diar)
    ubm_diar.segments = ubm_diar.segments[:1000]
    ubm_idmap = ubm_diar.id_map()

    if not os.path.isfile(f"{model_cfg['tmp_dir']}/model/ubm.h5"):
        ubm = sidekit.Mixture()
        logging.critical("Start UBM training with {} segments".format(len(ubm_idmap.leftids)))
        ubm.EM_split(fs_acoustic,
                     ubm_idmap,
                     model_cfg['model']['distrib_nb'],
                     num_thread=model_cfg['nb_thread'],
                     save_partial=False)
        ubm.write(f"{model_cfg['tmp_dir']}/model/ubm.h5")
    else:
        logging.critical("Load UBM from file")
        ubm = sidekit.Mixture(f"{model_cfg['tmp_dir']}/model/ubm.h5")
    logging.critical("\t Done")

    ######################################
    # Train TV
    ######################################
    if not os.path.isfile(f"{model_cfg['tmp_dir']}/model/TV.h5"):
        long_seg_diar = copy.deepcopy(train_diar)
        short_seg_diar = long_seg_diar.filter("duration", "<", 100)
        long_seg_diar = long_seg_diar.filter("duration", ">=", 100)
        tv_idmap = long_seg_diar.id_map()

        # Accumulate sufficient statistics for the training data
        logging.critical("Create StatServer with {} segments".format(len(tv_idmap.leftids)) )
        tv_stat = sidekit.StatServer(tv_idmap,
            distrib_nb=ubm.get_distrib_nb(),
            feature_size=model_cfg['model']['feature_size'] * 3
            )

        tv_stat.accumulate_stat(ubm=ubm,
            feature_server=fs_acoustic,
            seg_indices=range(tv_stat.segset.shape[0]),
            num_thread=model_cfg['nb_thread']
            )

        tv_stat = add_show_in_id(tv_stat)
        tv_stat.write(f"{model_cfg['tmp_dir']}/model/tv_stat.h5")
        logging.critical("\t Done")

        # Sufficient statistics are passed to the FactorAnalyser via an HDF5 file handler that stays in memory
        tv_stat_h5f = h5py.File(f'{model_cfg["tmp_dir"]}/tmp_tv_stat.h5', 'a', backing_store=False, driver='core')
        tv_stat.to_hdf5(tv_stat_h5f)
        #tv_stat.write(f'{model_cfg["tmp_dir"]}/tv_stat.h5', prefix='', mode='w')
        tv_fa = sidekit.FactorAnalyser()
        logging.critical("Start TV training with {} segments".format(len(tv_stat.modelset)))
        tv_fa.total_variability(tv_stat_h5f,
                                ubm,
                                tv_rank=model_cfg['model']['vectors']['size'],
                                nb_iter=10,
                                min_div=True,
                                tv_init=None,
                                batch_size=300,
                                save_init=False,
                                output_file_name=None,
                                num_thread=model_cfg['nb_thread'])
        tv_fa.write(f"{model_cfg['tmp_dir']}/model/TV.h5")
    else:
        logging.critical("Load TV from file")
        tv_fa = sidekit.FactorAnalyser(f"{model_cfg['tmp_dir']}/model/TV.h5")
    logging.critical("\t Done")

    ######################################
    # Extract I-vectors and train PLDA
    ######################################
    if not os.path.isfile(f"{model_cfg['tmp_dir']}/model/training_ivectors.h5"):

        # Accumulate sufficient statistics for the short segment from training data
        short_seg_idmap = short_seg_diar.id_map()
        short_seg_stat = sidekit.StatServer(short_seg_idmap,
            distrib_nb=ubm.get_distrib_nb(),
            feature_size=feature_size * 3
            )
        logging.critical("Start computing short segment statistics: {}".format(len(short_seg_stat.modelset)))
        short_seg_stat.accumulate_stat(ubm=ubm,
                                       feature_server=fs_acoustic,
                                       seg_indices=range(short_seg_stat.segset.shape[0]),
                                       num_thread=model_cfg['nb_thread']
                                       )
        logging.critical("short iv stat ok")
        # Rename in order to get one i-vector per segment
        #tv_stat = add_show_in_id(tv_stat)
        short_seg_stat = add_show_in_id(short_seg_stat)

        short_seg_stat.write(f"{model_cfg['tmp_dir']}/short_seg_stat.h5")



        logging.critical("Start extracting i-vectors")
        long_iv = tv_fa.extract_ivectors(ubm, f"{model_cfg['tmp_dir']}/model/tv_stat.h5", num_thread=model_cfg['nb_thread'])
        short_iv = tv_fa.extract_ivectors(ubm, f"{model_cfg['tmp_dir']}/model/short_seg_stat.h5",  num_thread=model_cfg['nb_thread'])

        current_iv = long_iv.merge(long_iv, short_iv)

        #long_iv.write(f"{model_cfg['tmp_dir']}/long_ivectors.h5")
        #short_iv.write(f"{model_cfg['tmp_dir']}/short_ivectors.h5")
        current_iv.write(f"{model_cfg['tmp_dir']}/model/training_ivectors.h5")
    else:
        logging.critical("Load IV from file")
        current_iv = sidekit.StatServer(f"{model_cfg['tmp_dir']}/model/training_ivectors.h5")
    logging.critical("\tI-vectors OK")


    # After taking the mean per speaker per file, we rename the clusters in order
    # to get one mean i-vector per speaker per file but same cluster_ID across shows
    current_iv_mean = current_iv.mean_stat_per_model()
    current_iv_mean = remove_show_from_id(current_iv_mean)

    # Normalize i-vectors
    norm_iv = copy.deepcopy(current_iv_mean)
    norm_mean, norm_cov = norm_iv.estimate_spectral_norm_stat1(1, 'sphNorm')
    norm_iv, n_recc, _ = keep_recurring_speakers(norm_iv, rank_F=iv_size, occ_number=2, filter_by_name=True)

    norm_iv.spectral_norm_stat1(norm_mean, norm_cov)
    logging.critical("start PLDA training")
    #plda_fa = sidekit.FactorAnalyser() Train  PLDA
    plda_fa = sidekit.FactorAnalyser()
    plda_fa.plda(norm_iv,
            rank_f=model_cfg['model']['plda_rank'],
            nb_iter=10,
            scaling_factor=1.,
            output_file_name=f"{model_cfg['tmp_dir']}/model/plda",
            save_partial=False)
    logging.critical("PLDA ok")

    model_iv = ModelIV()
    model_iv.ubm = ubm
    model_iv.tv = tv_fa.F
    model_iv.tv_mean = tv_fa.mean
    model_iv.tv_sigma = tv_fa.Sigma
    model_iv.sn_mean = norm_mean
    model_iv.sn_cov = norm_cov
    model_iv.plda_mean = plda_fa.mean
    model_iv.plda_f = plda_fa.F
    model_iv.plda_g = None
    model_iv.plda_sigma = plda_fa.Sigma
    model_iv.ivectors = current_iv
    model_iv.scores = None
    model_iv.nb_thread = 1

    model = dict()
    model["model_iv"] = model_iv
    model["global_diar"] = train_diar
    pickle.dump(model, open(f"{model_cfg['tmp_dir']}/model/model_allies_baseline.p", "wb" ))

    return model, model_cfg


def lium_xv_initial_training(model_cfg, train_data):
    """

    """

    if not os.path.isdir(f"{model_cfg['tmp_dir']}/model"):
        os.makedirs(f"{model_cfg['tmp_dir']}/model")

    train_diar = s4d.diar.Diar()
    df_data = dict()
    df_data['speaker_idx'] = []
    df_data['database'] = []
    df_data['speaker_id'] = []
    df_data['start'] = []
    df_data['duration'] = []
    df_data['file_id'] = []
    df_data['gender'] = []
    train_diar.add_attribut(new_attribut='gender', default='U')
    name_dict = {}

    # iterate on train_data
    spk_idx = 0
    speaker_idx = dict()
    for idx, (show, file_info, uem, ref, filename) in enumerate(train_data):

        # Use the ref info to fill the Diar object
        # fill a pandas.DataFrame at the sale tim to select data for X-vector extractor training
        for spk, start, end in zip(ref.speaker, ref.start_time, ref.end_time):
            if spk not in speaker_idx:
                speaker_idx[spk] = spk_idx
                spk_idx += 1

            train_diar.append(show=show,
                              cluster= spk,
                              start=int(round(float(start) * 100.)),
                              stop=int(round(float(end) * 100.))
                              #stop=int(round(float(end) * 100.)) + int(round(float(start) * 100.))
                              )
            df_data["speaker_idx"].append(speaker_idx[spk])
            df_data["speaker_id"].append(spk)
            df_data["start"].append(start)
            df_data["duration"].append(end)
            df_data["file_id"].append(show)
            df_data["database"].append("allies_train")
            df_data["gender"].append('u')

    print(f'Training data contains:')
    print("\t" + f'{len(train_diar.unique("show"))} files')
    print("\t" + f'{len(train_diar.unique("cluster"))} speakers')
    print("\t" + f'{len(train_diar)} segments')

    df = pandas.DataFrame(df_data)
    df = df.loc[df['duration'] > model_cfg['model']['vectors']['xvectors']['duration']]
    vc = df.speaker_idx.value_counts()
    selected_df = df[df.speaker_idx.isin(vc.index[vc.values >= model_cfg['model']['vectors']['xvectors']['min_nb_sessions']])]
    speaker_new_idx = dict(zip(list(selected_df.speaker_idx.unique()), numpy.arange(len(list(selected_df.speaker_idx.unique())))))
    selected_df = selected_df.reset_index(drop=True)
    for idx in range(len(selected_df)):
        selected_df.at[idx, 'speaker_idx'] = speaker_new_idx[selected_df.at[idx, 'speaker_idx']]
    selected_df.to_csv("xvector_dev.csv", index=False)

    training_speaker_nb = len(selected_df.speaker_idx.unique())
    print(f"Train Xtractor with {training_speaker_nb} speakers, from {len(selected_df)} sessions")

    #training_speaker_nb = 15
    model_cfg["model"]["speaker_nb"] = training_speaker_nb
    #model_cfg["model"]["speaker_nb"] = 111
    print("speaker_nb")
    if not os.path.isfile(model_cfg['tmp_dir']+"/model/best_xtractor.pt"):
        sidekit.nnet.xtrain(speaker_number=training_speaker_nb,
                            dataset_yaml=model_cfg['model']['vectors']['xvectors']['db_yaml'],
                            epochs=model_cfg['model']['vectors']['xvectors']['epochs'],
                            lr=model_cfg['model']['vectors']['xvectors']['lr'],
                            model_yaml=model_cfg['model']['vectors']['xvectors']['xtractor_yaml'],
                            model_name=None,
                            loss=model_cfg['model']['vectors']['xvectors']["loss"],
                            tmp_model_name=model_cfg['tmp_dir']+"/model/tmp_xtractor",
                            patience=50,
                            opt='adam',
                            best_model_name=model_cfg['tmp_dir']+"/model/best_xtractor",
                            multi_gpu=True,
                            num_thread=model_cfg['nb_thread'])

    # Extract x-vectors
    print("Load Training IdMap and extract x-vectors")
    train_diar = train_diar.filter("duration", ">=", 100)

    #############
    # Select speakers for PLDA training
    # Keep only sessions longer than model_cfg['model']['vectors']['xvectors']['duration']
    # And speakers having at least model_cfg['model']['vectors']['xvectors']['min_nb_sessions'] sessions
    #############
    train_diar = train_diar.filter("duration", ">=", int(model_cfg['model']['vectors']['xvectors']['duration'] * 100))

    training_idmap = train_diar.id_map()
    training_idmap.start = training_idmap.start * 160
    training_idmap.stop = training_idmap.stop * 160

    if not os.path.isfile(f"{model_cfg['tmp_dir']}/model/training_xv.h5"):
        training_xv = sidekit.nnet.xvector.extract_embeddings(idmap_name=training_idmap,
                                                              speaker_number=training_speaker_nb,
                                                              model_filename=model_cfg['tmp_dir']+"/model/best_xtractor.pt",
                                                              model_yaml=model_cfg['model']['vectors']['xvectors']['xtractor_yaml'],
                                                              data_root_name=train_data.root_folder + '/wav/',
                                                              device=torch.device("cuda:0"),
                                                              file_extension="wav",
                                                              transform_pipeline="MFCC,CMVN")
        training_xv.write(f"{model_cfg['tmp_dir']}/model/training_xv.h5") 

    training_xv = sidekit.StatServer(f"{model_cfg['tmp_dir']}/model/training_xv.h5")
    # Compute normalization parameters
    training_xv, n_recc, _ = keep_recurring_speakers(training_xv,
                                                     rank_F=model_cfg['model']['vectors']['size'],
                                                     occ_number=model_cfg['model']['vectors']['xvectors']['min_nb_sessions'],
                                                     filter_by_name=True)
    training_xv = add_show_in_id(training_xv)
    training_xv_mean = training_xv.mean_stat_per_model()
    training_xv_mean = remove_show_from_id(training_xv_mean)

    norm_xv = copy.deepcopy(training_xv_mean)
    norm_mean, norm_cov = norm_xv.estimate_spectral_norm_stat1(1, 'sphNorm')
    # Train PLDA
    print(f"Train PLDA model with {norm_xv.modelset.shape[0]} sessions from {len(set(norm_xv.modelset))} speakers")
    norm_xv.spectral_norm_stat1(norm_mean, norm_cov)
    plda_fa = sidekit.FactorAnalyser()
    plda_fa.plda(norm_xv,
                 rank_f=model_cfg['model']['plda_rank'],
                 nb_iter=10,
                 scaling_factor=1.,
                 output_file_name=None,
                 save_partial=False)


    model_iv = s4d.ModelIV()
    model_iv.sn_mean = norm_mean
    model_iv.sn_cov = norm_cov
    model_iv.plda_mean = plda_fa.mean
    model_iv.plda_f = plda_fa.F
    model_iv.plda_g = None
    model_iv.plda_sigma = plda_fa.Sigma
    model_iv.ivectors = training_xv
    model_iv.scores = None
    model_iv.nb_thread = 1

    model = dict()
    model["model_iv"] = model_iv
    model["global_diar"] = train_diar
    pickle.dump(model, open(f"{model_cfg['tmp_dir']}/model/model_allies_baseline.p", "wb" ))

    return model, model_cfg


def extract_vectors(model, model_cfg, current_diar, root_folder):
    """

    """
    if model_cfg["model"]["type"] == "lium_iv":
        # Load features

        # Extract i-vectors
        pass
    elif model_cfg["model"]["type"] == "lium_xv":
        # extract x-vectors
        current_im = current_diar.id_map()
        current_im.start = current_im.start * 160 
        current_im.stop = current_im.stop * 160 
        current_vec = sidekit.nnet.xvector.extract_embeddings(idmap_name=current_im,
                                                              speaker_number=model_cfg["model"]["speaker_nb"],
                                                              model_filename=model_cfg['tmp_dir']+"/model/best_xtractor.pt",
                                                              model_yaml=model_cfg['model']['vectors']['xvectors']['xtractor_yaml'],
                                                              data_root_name=root_folder + '/wav/',
                                                              device=torch.device("cuda:0"),
                                                              file_extension="wav",
                                                              transform_pipeline="MFCC,CMVN",
                                                              num_thread=10)

    return current_vec

def allies_init_seg(th_W, model, system_config, show, file_info, filename, root_folder):
    """

    """
    if isinstance(system_config, str):
        with open(system_config, 'r') as fh:
            model_cfg = yaml.load(fh, Loader=yaml.FullLoader)
    else:
        model_cfg = system_config

    if not os.path.isdir(f"{model_cfg['tmp_dir']}/seg/first/"):
        os.makedirs(f"{model_cfg['tmp_dir']}/seg/first/")
    if not os.path.isdir(f"{model_cfg['tmp_dir']}/seg/within/"):
        os.makedirs(f"{model_cfg['tmp_dir']}/seg/within/")

    # Perform first seg if it doesn't exist
    if not os.path.isfile(f"{model_cfg['tmp_dir']}/seg/first/{show}.seg"):
        # perform first seg
        if model_cfg["first_seg"]["type"] == "bic":

            if not os.path.isfile(f"{model_cfg['tmp_dir']}/feat/{show}.h5"):
                # Load MFCC
                fe = sidekit.FeaturesExtractor(audio_filename_structure="",
                                               feature_filename_structure="",
                                               sampling_frequency=16000,
                                               lower_frequency=133.3333,
                                               higher_frequency=6855.4976,
                                               filter_bank="log",
                                               filter_bank_size=40,
                                               window_size=0.025,
                                               shift=0.01,
                                               ceps_number=13,
                                               pre_emphasis=0.97,
                                               keep_all_features=True,
                                               vad='percentil',
                                               save_param=["energy", "cep", "vad"],
                                               compressed='none')

                print("Extract features")
                fe.save(show,
                        channel=0,
                        input_audio_filename=filename,
                        output_feature_filename=f"{model_cfg['tmp_dir']}/feat/{show}.h5")

            fs_seg = sidekit.FeaturesServer(feature_filename_structure=f"{model_cfg['tmp_dir']}/feat/{{}}.h5",
                                            dataset_list = ['energy', 'cep'],
                                            keep_all_features=True,
                                            mask='[1-13]',
                                            delta=False,
                                            double_delta=False)

            print("Load features'")
            cep, _= fs_seg.load(show)

            init_diar = s4d.segmentation.init_seg(cep, show)
            current_diar = s4d.segmentation.segmentation(cep, init_diar)

            if model_cfg["first_seg"]["bic_lin"]:
               current_diar = s4d.segmentation.bic_linear(cep, current_diar, model_cfg['first_seg']['thr_l'], sr=False)

            if model_cfg["first_seg"]["hac_bic"]:
                hac = s4d.clustering.hac_bic.HAC_BIC(cep, current_diar, model_cfg['first_seg']['thr_h'], sr=False)
                current_diar = hac.perform()

            if model_cfg["first_seg"]["viterbi"]:
                current_diar = s4d.viterbi.viterbi_decoding(cep, current_diar, model_cfg['first_seg']['thr_vit'])

        else:
            #slidding window
            nfo = soundfile.info(filename)
            # Get beginning and end of the file
            start = 0.0
            end = float(nfo.frames) / 16000.

            current_diar = s4d.Diar()
            current_diar.add_attribut(new_attribut='gender', default='U')
            s = 0.0
            e = float(model_cfg["first_seg"]["sliding_window"])
            idx = 0
            while e < end:
                current_diar.append(show=show,
                                    cluster=f"tmp{idx}",
                                    start=int(s * 100),
                                    stop=int((s + e) * 100),
                                    gender='U')

                idx += 1
                s += float(model_cfg["first_seg"]["sliding_shift"])
                e += float(model_cfg["first_seg"]["sliding_window"])

        s4d.Diar.write_seg(f"{model_cfg['tmp_dir']}/seg/first/{show}.seg", current_diar)

    # load the first seg
    current_diar = s4d.Diar.read_seg(f"{model_cfg['tmp_dir']}/seg/first/{show}.seg")

    # Extract segment representation (i-vectors or x-vectors)
    if not os.path.isfile(f"{model_cfg['tmp_dir']}/seg/first/{show}_xv.h5"):
        current_vec = extract_vectors(model, model_cfg, current_diar, root_folder)
        current_vec.write(f"{model_cfg['tmp_dir']}/seg/first/{show}_xv.h5")
    else:
        current_vec = sidekit.StatServer(f"{model_cfg['tmp_dir']}/seg/first/{show}_xv.h5")

    scores = None
    #backup_current_diar = copy.deepcopy(current_diar)
    #backup_current_vec = copy.deepcopy(current_vec)
    if not isinstance(th_W, list):
        th_W = [th_W,]

    for th in th_W:
        
        backup_current_diar = copy.deepcopy(current_diar)
        backup_current_vec = copy.deepcopy(current_vec)

        if not os.path.isdir(f"{model_cfg['tmp_dir']}/seg/within_{th}/"):
            os.makedirs(f"{model_cfg['tmp_dir']}/seg/within_{th}/")

        # Perform HAC-xv if necessary
        backup_current_diar, backup_current_vec, scores = perform_iv_seg(model["model_iv"].sn_mean,
                                                           model["model_iv"].sn_cov,
                                                           model["model_iv"].plda_mean,
                                                           model["model_iv"].plda_f,
                                                           model["model_iv"].plda_sigma,
                                                           th,
                                                           backup_current_diar,
                                                           backup_current_vec,
                                                           model_cfg["within_show"]["hac_method"],
                                                           scores)
        s4d.Diar.write_seg(f"{model_cfg['tmp_dir']}/seg/within_{th}/{show}.seg", backup_current_diar)
        allies_write_diar(backup_current_diar, f"{model_cfg['tmp_dir']}/seg/within_{th}/{show}.mdtm")
        backup_current_vec.write(f"{model_cfg['tmp_dir']}/seg/within_{th}/{show}_vec.h5")

    return current_diar, current_vec, scores


def allies_within_show_HAL(current_diar,
                           current_vec,
                           scores,
                           th,
                           model_cfg,
                           model,
                           file_info,
                           user,
                           uem,
                           ref):

    # get correction from the user
    message_to_user = MessageToUser(file_info,
                                    s4d_to_allies(current_diar),
                                    Request('same', 0., 1.))
    hal, answer = user.validate(message_to_user)
    answer.time_1 = int(answer.time_1 * 100)
    answer.time_2 = int(answer.time_2 * 100)

    # Get the index of the corrected segments in the diarization
    for idx, seg in enumerate(current_diar.segments):
        if seg['start'] <= answer.time_1 <= seg["stop"]:
            idx_1 = idx
            cluster_1 = current_diar[idx_1]['cluster']
        if seg["start"] <= answer.time_2 <= seg["stop"]:
            idx_2 = idx
            cluster_2 = current_diar[idx_2]['cluster']

    # apoply corretion and produce the new segmentation
    # Only correct when the two segments have to be grouped
    if answer.response_type == "same" and answer.answer:
       current_diar = apply_correction(current_diar, scores, idx_1, idx_2)

    return current_diar, current_vec

def allies_cross_show_clustering(show_idx, model, model_cfg, current_diar, current_vec, th_x):
    """

    """
    if show_idx == 0:
        model["previous_vec"] = copy.deepcopy(current_vec)
        model["previous_diar"] = current_diar

    else:
        #previous_diar.segments += within_diar.segments
        # Apply cross-show clustering
        previous_vec, previous_diar, current_diar = cross_show(model=model,
                                                               model_cfg=model_cfg,
                                                               previous_iv=model["previous_vec"],
                                                               previous_diar=model["previous_diar"],
                                                               within_iv=current_vec,
                                                               within_diar=current_diar,
                                                               th_x=th_x,
                                                               incremental=True)
        model["previous_vec"] = previous_vec
        model["previous_diar"] = previous_diar

    return model, current_diar


def allies_write_diar(current_diar, filename):
    """

    """
    for idx, seg in enumerate(current_diar):
        current_diar.segments[idx]['start'] = float(seg['start']) / 100.
        current_diar.segments[idx]['stop'] = float(seg['stop']) / 100.


    current_diar.sort(['show', 'start'])
    with open(filename, 'w', encoding="utf8") as fic:
        for line in s4d.Diar.to_string_seg(current_diar, time_float=True):
            fic.write(line)





#!/usr/bin/

import os
import numpy as np
import wave
import struct
from collections import namedtuple
from .user_simulation import FileInfo
from .user_simulation import Reference
from .user_simulation import UEM

def get_key(filename):
    if filename[0:3] == "199" or filename[0:3] == "200":
        return filename[0:8] + "." + filename[9:13] + "." + filename[19:]
    l = len(filename)
    if filename[l - 17 : l - 14] == "201":
        return (
            filename[l - 17 : l - 13]
            + filename[l - 12 : l - 10]
            + filename[l - 9 : l - 7]
            + "."
            + filename[l - 6 : l - 2]
            + "."
            + filename[0 : l - 18]
        )
    if filename[14:18] == "2010":
        if filename[27:28] == "_":
            return filename[14:22] + "." + filename[23:27] + "." + filename[28:]
        else:
            return filename[14:22] + "." + filename[23:27] + "." + filename[30:]
    return "???"


def get_file_list(path):
    pairs = []
    for name in os.listdir(path):
        if not name.startswith('.'):
            name = name[:-5]
            k = get_key(name)
            pairs.append([k, name])
    pairs.sort(key=lambda p: p[0])
    return pairs


def hashrand(a):
    a = (a + 0x7ED55D16) + (a << 12)
    a = (a ^ 0xC761C23C) ^ (a >> 19)
    a = (a + 0x165667B1) + (a << 5)
    a = (a + 0xD3A2646C) ^ (a << 9)
    a = (a + 0xFD7046C5) + (a << 3)
    a = (a ^ 0xB55A4F09) ^ (a >> 16)
    a = a & 0xFFFFFFFF
    a = a / 0xFFFFFFFF
    return a


class Database:
    """
    Object that load data from the ALLIES database and create an iterator that will
    yield requested information to process a file:
    file_info, uem, speakers

    speakers is the reference from MDTM
    """
    def __init__(self, root_folder, parameters):

        self.index = 0
        self.root_folder = root_folder

        self.fl = get_file_list(root_folder + "/mdtm")

        Entry = namedtuple(
            # 'Entry', ['root_folder', 'filename', 'file_info', 'speech', 'uem', 'speakers', 'llmode']
            "Entry",
            ["root_folder", "filename", "file_info", "speech", "uem", "speakers"],
        )

        self.ents = []
        for idx, f in enumerate(self.fl):
            if ("start_date" in parameters) and f[0] < parameters["start_date"]:
                continue
            if ("end_date" in parameters) and f[0] >= parameters["end_date"]:
                continue
            llmode = "none"
            if "interactive_ratio" in parameters:
                if float(parameters["interactive_ratio"]) > hashrand(idx):
                    llmode = "interactive"
                else:
                    llmode = "active"
            self.ents.append(
                Entry(
                    root_folder,
                    f[1],
                    FileInfo(f[0], llmode, f[0][0:13]),
                    [f[0]],
                    [f[0]],
                    [f[0]],
                )
            )
        self.number_of_files = len(self.ents)

    def __iter__(self):
        return self

    def __next__(self):
        """
        Return a tuple including 
            file_info, uem, speakers
        :return: 
        """
        if self.index < self.number_of_files:

            # GET filename
            filename = self.ents[self.index].filename
            complete_filename = self.root_folder + "/wav/" + self.ents[self.index].filename + ".wav"

            # GET file_info
            file_info = self.ents[self.index].file_info

            # GET UEM
            st = []
            en = []
            for l in open(self.root_folder + "/uem/" + self.ents[self.index].filename + ".uem", "r"):
                e = l.split()
                st.append(e[2])
                en.append(e[3])
            #uem = {"start_time": np.cast["float64"](st), "end_time": np.cast["float64"](en)}
            uem = UEM(np.cast["float64"](st), np.cast["float64"](en))

            # GET Speakers
            spk = []
            ref_st = []
            ref_en = []
            for l in open(self.root_folder + "/mdtm/" + self.ents[self.index].filename + ".mdtm", "r"):
                e = l.split()
                ref_st.append(np.cast["float64"](round(float(e[2]), 3)))
                ref_en.append(np.cast["float64"](round(float(e[2]) + float(e[3]), 3)))
                spk.append(e[7])
            speakers = Reference(spk, ref_st, ref_en)
            self.index += 1
            return filename, file_info, uem, speakers, complete_filename

        else:
            raise StopIteration()
